/*

Example participants
{
	"10000090": {
		"company_id": 0,
		"first_name": "First-090",
		"last_name": "Xlast-090",
		"email": "email-090@test.com",
		"phone_number": "0000000090",
		"presence": 1,
		"status": "Available",
		"emails": null,
		"ImageUrl": "http://i1286.photobucket.com/albums/a614/eqv7/numbers/090.jpg",
		"ImageUrlSmall": "http://robohash.org/First-090 Xlast-090.png?set=set1",
		"ImageUrlLarge": "http://robohash.org/First-090 Xlast-090.png?set=set1",
		"title": "Title-090",
		"is_confirmed": false,
		"author_email": null,
		"fav_employees": null,
		"id": 10000090,
		"date_created": "0001-01-01T00:00:00",
		"date_modified": null
	}
}

presence :
0	offline		FromRGB (152, 152, 152)
1	available	FromRGB (52,80,109)
2	away		FromRGB (248,177,51)
3	busy		FromRGB (207,72,38)


Example message
	f => MessageType
		Regular = 0,
		System = 1,
		Chat2Email = 2,
		VoipCall = 3,
		VoipCallEnd = 4,
		VoipCallMissed = 5,
		ConfCallInitiated = 6,
		ConfCallStarted = 7,
		ConfCallEnded = 8,
		GroupCreated = 9
	aid => AuthorId
	t => Text
	ds => DateSent

{
	"f": 0,
	"aid": 366,
	"t": "her",
	"id": "q0h_o-Ved0uuibZGQuJtOw",
	"ds": "2013-09-18T17:23:49.023Z"
},
{
	"f": 3,
	"aid": 356,
	"tid": 366,
	"t": "Calling you",
	"id": "offer-356-0d32a8cb-84e5-47dc-8f81-30a7da0aabbc",
	"ds": "2013-09-18T01:00:23.459Z"
}

*/

var Chat = (function (){
	var me = {};
	var _currentUserId;
	var _staticUri = "http://static.getclementine.com";
	var _participants = {};
	var _numParticipants = 0;
	var _messages = [];
	var _supportedMessageTypes = [0,2];	// Regular and Chat2Email
	var _onBottom = true;
	var _unreadMessageCount = 0;
	var _unreadMessageClass = "";
	var _curDateBlock = null;
	var _curDateBlockLastMessage = null;
	var _dataKeys = {
						dateObject : 'date'
					};
	var _separator = "  &nbsp;&#149;&nbsp;  ";

	me.init = function () {
		window.onscroll = _onScroll;
		window.addEventListener("touchmove", _onScroll, false);	// ios scroll change detection

		$('#bottombar').hide();
		$('#bottombar').click(_bottombarClicked);

		$(document).on('inview', '.message.new', function() {
			$(this).animate({backgroundColor: 'transparent'}, 2000, function () {
				$(this).removeClass('new');

			});
		});

		var typingDiv = $('<div id="typing" style="visibility:hidden;"><div class="pics"></div><div class="text">typing...</div></div>');
		$('#bottomDiv').append(typingDiv);
	};

	/**************************************** External API ****************************************/

	me.SetCurrentUserEnvironment = function (id, staticUri) {
		_currentUserId = id;
		_staticUri = staticUri;
	};

	me.SetSupportedMessageTypes = function (types) {
		_supportedMessageTypes = types;
	};

	me.LoadParticipants = function (participants) {
		_participants = participants;
		_numParticipants = Object.keys(_participants).length;
	};

	me.UpdateParticipant = function (participant) {
		var prevParticipant = _participants[participant.id];
		if (prevParticipant != null) {
			// update existing participant
			var presenceChanged = (prevParticipant.presence != participant.presence);
			var memberJoined = (!prevParticipant.is_confirmed && participant.is_confirmed);
			if (presenceChanged || memberJoined) {
				// presence has been changed
				var pics = $('.pic[data-userId=' + participant.id +']');
				if (pics.length > 0) {
					pics.removeClass('nonUser').removeClass('offline')
						.removeClass('available').removeClass('away')
						.removeClass('busy').removeClass('custom').addClass(_getPresenceClass(participant));
				}
			}
		}
		_participants[participant.id] = participant;
	}

	// Initial loading of existing chat messages
	me.LoadMessages = function (messages, skipAnimate) {
		var html = $('#chat');
		for (var i = 0; i < messages.length; i++) {
			_insertMessage(messages[i]);
		}
		if (skipAnimate) {
			$(window).scrollTop($('#bottomDiv').offset().top);
            Cl.API.PageRendered();
			return;
		}

		$('#bottomDiv').goTo(function () {
			Cl.API.PageRendered();
		});
	};

	/* After adding messages it will put the */
	me.AddPreviousMessages = function (messages) {
		if (messages.length == 0)
			return;

		// make #chat invisible while adding messages
		$('#chat').css('visibility', 'hidden');

		var html;
		var existingMsgs = $('.message');
		var topMsg = (existingMsgs.length > 0) ? $(existingMsgs[0]) : $('#bottomDiv');
		var curY = window.scrollY;
		messages.forEach(function (msg) {
			_insertMessage(msg);
		});
		// move scroll without animation
		// keep the position of current view
		$('html, body').animate(
			{ scrollTop: topMsg.offset().top},
			0,
			'swing',
			function (){
				// make #chat visible when adding message and scroll to right place is done
				$('#chat').css('visibility', 'visible');
			});
	};

	me.AddMessage = function (message, skipAnimate) {
		var animate = _isWatchingLastMessage();

		var pair = _insertMessage(message);
		var isNew = pair[0];
		var html = pair[1];
		if (skipAnimate)
			return html;
		if (animate) {
			_scrollToBottom();
		} else {
			if (isNew) {
				_unreadMessageCount++;
				$(html).addClass('new');
				_unreadMessageClass = message.id;
			}
			if (_unreadMessageCount > 0) {
				var text = _unreadMessageCount.toString();
				text += (_unreadMessageCount === 1) ? ' new message' : ' new messages';
				$('#bottombar').html(text).show();
			}
		}
		return html;
	};

	me.AddPendingMessage = function (message) {
		var pair = _insertMessage(message, true);
		var html = pair[1];
		_scrollToBottom();
		$(html).addClass('pending');
		$(html).prepend('<div class="spinner"/>');
		// if message is not sent within 1 seconds, treat it as failed
		waitingForPendingMessageToGoThrough(message.id);
	};

	var _typingUsers = {};
	//var _typingTimer = null;

	me.Typing = function (participantId, isTyping) {
		var typingPicDiv = $('#bottomDiv #typing .pics');
		if (isTyping) {
			_typingUsers[participantId] = new Date();
			if ($('#typing [data-userId=' + participantId +']').length == 0) {
				var profileDiv = $(_getProfileImageDiv(participantId));
				typingPicDiv.prepend(profileDiv);
				$('#typing').css('visibility','visible');
			}
			//if (_typingTimer == null) {
			//	_typingTimer = setInterval(_typingIndicatorTimer, 1000);
			//}
		} else {
			if (participantId in _typingUsers) {
				delete _typingUsers[participantId];
				$('#typing [data-userId=' + participantId +']').remove();
				_typingIndicatorCleanup();
			}
		}
	};

	//function _typingIndicatorTimer() {
	//	var curTime = new Date();
	//	for(var id in _typingUsers) {
	//		if ((curTime - _typingUsers[id]) > 2000) {
	//			delete _typingUsers[id];
	//			$('#typing [data-userId=' + id +']').remove();
	//		}
	//	}
	//	_typingIndicatorCleanup();
	//}

	function _typingIndicatorCleanup() {
		var typing = $('#typing .pic');
		if (typing.length == 0) {
			$('#typing').css('visibility','hidden');
			//if (_typingTimer != null) {
			//	clearInterval(_typingTimer);
			//	_typingTimer = null;
			//}
		}
	}

	me.Clear = function () {
		$('#chat').html('');
		_curDateBlock = null;
		_curDateBlockLastMessage = null;
	};

/**************************************** Functional Codes ****************************************/

	function _isWatchingLastMessage() {
/* 		Cl.Log(window.innerHeight, window.scrollY, window.innerHeight + window.scrollY, document.body.offsetHeight, ((window.innerHeight + window.scrollY) >= document.body.offsetHeight) ? "bottom" : "not bottom"); */
		var threshold = 30;	// if user is looking at the threshold pixels above from the bottom of the page, consider it as looking at the last message
		return ((window.innerHeight + window.scrollY) >= document.body.offsetHeight - threshold);
	}


	function waitingForPendingMessageToGoThrough (id) {
		var html = $('.message.' + id);
		window.setTimeout(function() {
			if ($(html).hasClass('pending')) {
				$(html).addClass('error');
				$('.' + id + '.error .spinner').click(function(){
					$(html).removeClass('error');
					Cl.API.RetrySendMessage(id);
					waitingForPendingMessageToGoThrough (id);
				});
			}
		}, 3000);

	}

	function _onScroll() {
		if ((window.innerHeight + window.scrollY) >= document.body.offsetHeight) {
			_onBottom = true;
			// hide unread message div
			$('#bottombar').hide();
			_unreadMessageCount = 0;
		} else {
			_onBottom = false;
		}
/* 		Cl.Log(window.innerHeight, window.scrollY, window.innerHeight + window.scrollY, document.body.offsetHeight, _onBottom ? "bottom" : "not bottom"); */
	}

	function _bottombarClicked() {
		$('.' + _unreadMessageClass).find('.timeStamp').goTo();
	}

	function _scrollToBottom() {
		$('#bottomDiv').goTo();
	}

/****************************************  Only place to add message to Dom ****************************************/

	function _insertMessage(message, asLastMsg) {
		var isNew;

		// Make sure duplicate messages do not get added to the screen
		var existingMessage = $('.' + message.id);
		if (existingMessage.length > 0) {
			isNew = false;
			// if pending message existed, remove the pending class to confirm message delivery
			if (existingMessage.hasClass('pending')) {
				existingMessage.find('.timeStamp').html(_getTimeString(message.ds));
				var t = new Date(message.ds);
				existingMessage.data(_dataKeys.dateObject, t);
				existingMessage.removeClass('pending');

				// TODO: should re-order pending message if ordering is incorrect
				// check if current ordering is correct
				var nextMessages = existingMessage.nextAll('.message');
				// if there was a message after current message and that time is older than pending message,
				// we need to re-order pending message
				if (nextMessages.length > 0 && $(nextMessages[0]).data(_dataKeys.dateObject) < t) {
					existingMessage.remove();
					existingMessage = null;
				}
				// if not, remove it from DOM and re add it;
			}
			// else message is already on chat & does not need to be re-ordered, so ignore
			if (existingMessage != null) {
				return [isNew, existingMessage];
			}
		} else {
			isNew = true;
		}

		var msg = _renderMessage(message);
		// incase of unsupported message type, just return early
		if (msg.length == 0)
			return [isNew, msg];

		// if _curDateBlock does not exist or _curDateBlock's date is different than the msg's date, change _curDateBlock
		if (_curDateBlock == null || (!_isDateMatchWithDateBlock(_curDateBlock, msg.data(_dataKeys.dateObject)))) {
			_curDateBlockLastMessage = null;
			_curDateBlock = _findDateBlockForDate(msg.data(_dataKeys.dateObject));
			if (_curDateBlock != null) {
				var msgs = _curDateBlock.children('.message');
				_curDateBlockLastMessage = $(msgs[msgs.length - 1]);
			}
		}

		// if dateBlock for current date exist already
		if (_curDateBlock) {
			if (!_curDateBlockLastMessage) {
				_curDateBlock.append(msg);
				_curDateBlockLastMessage = msg;
				return [isNew, msg];
			}

			if (_curDateBlockLastMessage && _curDateBlockLastMessage.data(_dataKeys.dateObject) < msg.data(_dataKeys.dateObject)) {
				_curDateBlockLastMessage.after(msg);
				_curDateBlockLastMessage = msg;
				return [isNew, msg];
			}
			// message should be added to this block
			if (!asLastMsg) {
				var messages = _curDateBlock.children('.message');
				_curDateBlockLastMessage = $(messages[messages.length - 1]);
				for (var i = messages.length - 1; i >= 0 ; i--) {
					var curMsg = $(messages[i]);
					if (curMsg.data(_dataKeys.dateObject) < msg.data(_dataKeys.dateObject)) {
						curMsg.after(msg);
						return [isNew, msg];
					}
				}
				_curDateBlock.find('.metaData.timeline').after(msg);
			} else {
				_curDateBlockLastMessage = msg;
				_curDateBlock.append(msg);
			}

			return [isNew, msg];
		}

		// dateBlock with the same date
		var newBlock = _renderDateBlock(msg);
		if (newBlock.length == 0) {
			return [isNew, newBlock];
		}

		newBlock.append(msg);
		_insertDateBlock(newBlock);
		_curDateBlock = newBlock;
		_curDateBlockLastMessage = msg;

		// there was no dateBlock for this time
		return [isNew, msg];
	}

	function _insertDateBlock(block) {
		var dateBlocks = $('.dateBlock');
		for (var i = 0; i < dateBlocks.length; i++) {
			var curBlock = $(dateBlocks[i]);
			if (block.data(_dataKeys.dateObject) < curBlock.data(_dataKeys.dateObject)) {
				curBlock.before(block);
				return;
			}
		}
		$('#chat').append(block);
	}

	function _findDateBlockForDate(date) {
		var dateBlocks = $('.dateBlock');
		for (var i = dateBlocks.length - 1; i >= 0; i--) {
			var curBlock = $(dateBlocks[i]);
			if (_isDateMatchWithDateBlock(curBlock, date)) {
				return curBlock;
			}
		}
		return null;
	}

	function _isDateMatchWithDateBlock(block, date) {
		var t = block.data(_dataKeys.dateObject);
		return (t.getDate() == date.getDate() && t.getDay() == date.getDay() && t.getYear() == date.getYear());
	}

/**************************************** DOM Definition ****************************************/

	// msg : jquery object of new message div
	function _renderDateBlock(msg) {
		if (!msg.data(_dataKeys.dateObject)) {
			return $();
		}
		var date = msg.data(_dataKeys.dateObject);
		var html = $('<div class="dateBlock">' + '</div>');
		html.data(_dataKeys.dateObject, date);
		html.append(_renderTimeLine(date));
		return html;
	}

	// date : javascript Date object
	function _renderTimeLine(date) {
		var html=   '<div class="metaData timeline">' +
						'<span>' + _getDateString(date) + '</span>' +
					'</div>';
		return $(html);
	}


	// message : raw json message object
	function _renderMessage(message) {
		var type = 'other';
		if (message.aid == _currentUserId) {
			type = 'me';
		}

		if (_supportedMessageTypes.indexOf(message.f) == -1)
			return $();
//			return $('<div class="' + message.id + '"></div>');

		// Message Type : Regular & Chat2Email (regular bubble chat UI)
		if (message.f === 0 || message.f === 2) {
			// TODO: add Chat2Email icon in div
			var classForMessageOrigin = '';
			if (message.f === 2 && message.an != "Clementine Team") {
				classForMessageOrigin = ' email';
			}

			// if it is 1:1 chat with non-app user, we wanna show "sent via email" for my messages
			if (_numParticipants == 2 && type == 'me') {
				for(var id in _participants) {
					if (id != message.aid) {
						if (id && !_participants[id].is_confirmed && _participants[id].email.indexOf("team@clementine") > -1) {
							classForMessageOrigin = ' sentViaEmail';
						}
					}
				}
			}

			// Only show name for chats w/ more than 2 people
			senderDiv = '';
			if (_numParticipants > 2 && type != 'me') {
				senderDiv = '<div>' +
								'<div class="timeStamp">' + _getTimeString(message.ds) + '</div>' +
								'<div class="name" onclick="Cl.API.OpenContact(' + message.aid + ')">' + _getName(message.aid) + '</div>'
							'</div>';
			} else {
				senderDiv = '<div class="timeStamp">' + _getTimeString(message.ds) + '</div>';

			}

			var html = '<div class="message ' + type + ' ' + message.id + classForMessageOrigin + '">' +
							_getProfileImageDiv(message.aid) +
							'<div class="bubble ' + type + '">' +
								'<div class="chatText">' + Cl.htmlEcnoder.parseMessage(message.t) +
								senderDiv
								'</div>' +
							'</div>' +
						'</div>';
			return $(html).data(_dataKeys.dateObject, new Date(message.ds));
		// Message Type : System (No use case yet)
		} else if (message.f === 1) {
			var html =	'<div class="metaData bubble system message ' + message.id + '">' +
							'<span>' +
								message.t +
							'</span>' +
						'</div>';
			return $(html).data(_dataKeys.dateObject, new Date(message.ds));
		// Message Type : VoipCall & VoipCallEnd & VoipCallMissed (VoipCall related Meta Data)
		} else if (message.f === 3 || message.f === 4 || message.f === 5) {
			var msg = '';
			if (message.f === 3) {
				if (message.aid == _currentUserId) {
					msg = '<b>You</b> Calling';
				} else {
					msg = '<b>' + _getName(message.aid) + '</b>' + ' Calling';
				}
			} else if (message.f === 4) {
				msg = 'Call Ended';
			} else if (message.f === 5) {
				msg = 'Call Missed';
			}
			var html =	'<div class="metaData message voip ' + message.id + '">' +
							'<span>' +
								msg +
								'<div class="timeStamp">' + _separator + _getTimeString(message.ds) + '</div>' +
							'</span>' +
						'</div>';
			return $(html).data(_dataKeys.dateObject, new Date(message.ds));
		} else if (message.f === 6 || message.f === 7 || message.f === 8) {
			var msg = '';
			if (message.f === 6) {
				if (message.aid == _currentUserId) {
					msg = 'Call initiated by <b>You</b>';
				} else {
					msg = 'Call initiated by <b>' + _getName(message.aid) + '</b>';
				}
			} else if (message.f === 8) {
				var diff = new Date(message.ds) - new Date(message.st);
				msg = message.t + ' (' + _millisecondsToTime(diff) + ')';
			} else {
				msg = message.t;
			}
			var html =	'<div class="metaData message conf ' + message.id + '">' +
							'<span>' +
								msg +
								'<div class="timeStamp">' + _separator + _getTimeString(message.ds) + '</div>' +
							'</span>' +
						'</div>';
			return $(html).data(_dataKeys.dateObject, new Date(message.ds));
		} else if ([9,11,12,13,14].indexOf(message.f)>=0 && message.t) {
			var html =	'<div class="metaData message group ' + message.id + '">' +
							'<span>' +
								message.t +
								//'<div class="timeStamp">' + _separator + _getTimeString(message.ds) + '</div>' +
							'</span>' +
						'</div>';
			return $(html).data(_dataKeys.dateObject, new Date(message.ds));
		}
		return $();
	}

	function _getProfileImageDiv(participantId) {
		var participant = _participants[participantId];
		var imageDiv;
		if (_getProfileImageUrl(participant)) {
			imageDiv = '<img src="' + _getProfileImageUrl(participant) + '" onerror="this.src=\'img/default_profile.png\'"/>';
		} else {
			imageDiv = '<div class="initial">' + _getInitials(participant) + '</div>';
		}
		var html = 	'<div class="pic ' + _getPresenceClass(participant) + '" onclick="Cl.API.OpenContact(' + participantId + ')" data-userId="' + participantId + '">' +
						imageDiv +
						//'<div class="presenceDot"></div>' +
					'</div>';
		return html;
	}

/**************************************** Getter Methods ****************************************/

	function _getName(participantId) {
		var participant = _participants[participantId];

		if (!participant) {
			Cl.Log("Unable to find participant: " + participantId + " In dictionary " + JSON.stringify(_participants));
			return "????";
		}


		// Show domain for externals
		var domain = '';
		/*
		if (participant.company_id != _participants[_currentUserId].company_id) {
			domain = ' (' + _getEmailDomain(participant.email) + ')';
		}
		*/


		if (participant.first_name) {
			return participant.first_name + domain;
		}
		if (participant.last_name) {
			return participant.last_name + domain;
		}
		if (participant.email) {
			// TODO: maybe substring from beginning to '@' to make it shorter?
			return participant.email.substr(0, participant.email.indexOf('@'));
		}
	}

	function _getEmailDomain(email) {
		if (!email || email.indexOf('@') == -1)
			return '';
		return email.substr(email.indexOf('@') + 1);
	}

	function _getInitials(participant) {
		var str = '';
		if (!participant) {
			return '?';
		}
		if (participant.first_name) {
			str += participant.first_name[0];
		}
		if (participant.last_name) {
			str += participant.last_name[0];
		}
		return str;
	}

	function _getPresenceClass(participant) {
		var pres = '';
		if (!participant || !participant.is_confirmed) {
			return 'nonUser';
		} else {
			pres = 'user ';
		}

		if (!participant || participant.presence === 0) {
			return pres + 'offline';
		} else if (participant.presence === 1) {
			return pres + 'available';
		} else if (participant.presence === 2) {
			return pres + 'away';
		} else if (participant.presence === 3) {
			return pres + 'busy';
		} else {
			return pres + 'custom';
		}
	}

	function _getProfileImageUrl(participant) {
		if (!participant || !participant.image_hash) {
			return "img/default_profile.png";
		} else {
			var url = _staticUri + "/users/" + participant.id + "_" + participant.image_hash + ".small.jpg";
			Cl.Log("Url = " + url);
			return url;
		}

	}

	function _getDateAndTimeString(date) {
		date = new Date(date);
		return date.toLocaleDateString() + ' ' + date.toLocaleTimeString();
	}

	function _getTimeString(date) {
		date = new Date(date);
		var hour = date.getHours();
		var min = date.getMinutes();
		var ampm;
		// AM/PM handling
		if (hour < 12) {
			ampm = ' AM';
		} else {
			ampm = ' PM';
			if (hour > 12) {
				hour -= 12;
			}
		}
		// single digit minute handling
		if (min < 10) {
			min = '0' + min;
		}
		return hour + ':' + min + ampm;
//		return date.toLocaleTimeString();
	}

	function _getDateString(date) {
		date = new Date(date);
		return date.toLocaleDateString();
	}

	function _millisecondsToTime(s) {
		function addZ(n) {
			return (n<10? '0':'') + n;
		}

		var ms = s % 1000;
		s = (s - ms) / 1000;
		var secs = s % 60;
		s = (s - secs) / 60;
		var mins = s % 60;
		var hrs = (s - mins) / 60;

		return addZ(hrs) + ':' + addZ(mins) + ':' + addZ(secs);
	}

	return me;
}());

(function($) {
    $.fn.goTo = function(callback) {
		if (typeof(callback) != "function") {
			callback = null;
		}
    	if (this.length > 0) {
			$('html, body').animate({
				scrollTop: $(this).offset().top + 'px'
			}, 'fast', callback);
			return this;
		}
		if (callback) {
			callback();
		}
        return this; // for chaining...
    }
})(jQuery);
